/*
 * @(#)jdemoextapi.java    1.00   08/18/04
 *
 * Copyright (c) 2004++ Agilis Software LLC, Santa Clara, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 *
 * DISCLAIMER: this sample code is provided as-is.
 */

/*
 * This demo illustrates how to securely manage state in an
 * EasyLicenser-enabled application using the extended
 * API's for maintaining quota consumption, deferred-time-limited
 * licensing, deferred node locking, and other license-related 
 * application state.
 *
 * The demo is invoked in one of 3 modes:
 * Activation: this is equivalent to an installation or activation of
 * a protected application.
 * Execution: this is equivalent to an execution of a protected application
 * in the steady state.
 * Deactivation: this is equivalent to an uninstallation or explicit
 * deactivation of a protected application.
 * Running the demo once in execution mode is equivalent to your end 
 * customer running your license-protected application at their site 
 * and consuming a specified amount of quota at their site.  
 * The quota consumption is specified as a command line argument.
 *
 * Features illustrated with this demo:
 * 1. How to use the extended EasyLicenser API calls in conjunction with
 *    the key cookie mechanism to manage and update license state 
 *    including the built-in quota management and the extensibility 
 *    mechanism to manage application-specific state, as well as to
 *    illustrate deferred-time-limited licensing and deferred node
 *    locking.
 * 2. How to utilize the application-password security mechanism.
 * 3. How to protect against system clock changes without writing code.
 *
 * For examples on using custom key handlers for extensibility,
 * as well as using digital signatures and other anti-spoofing measures,
 * see the accompanying "jdemosecure.java" demo application.
 * For an example of using the core API to accomplish secure state
 * management, see the accompanying "jdemostate.java" demo application.
 *
 * Some important points to note about this demo:
 * 1. The code that is directly relevant to using the EasyLicenser API's
 *    is indicated as such in comments.  The rest of the code exists in
 *    order to have a functioning self-contained program.
 * 2. All code pertaining to reading the key from an EasyLicenser file is
 *    for functional purposes only.  It is not meant to suggest a specific
 *    technique for managing the key in your application.  Nor is it
 *    a part of the EasyLicenser product.  In particular, note that your
 *    actual application will typically have access to just the license key
 *    without any accompanying metadata about the key.
 * 3. The demo will work with keys generated with EasyLicenser 2.0 as well
 *    as prior versions.  The EasyLicenser 2.0 run time library is required.
 * 4. This demo performs automatic state management in a file that persists
 *    past its execution.  Therefore, each successive run will have a 
 *    different outcome and eventually you should encounter an "exceeded
 *    quota" error if you supplied a quota limited key, or a "license
 *    expired" error if you specified an auxiliary trial-generator key
 *    (even if your key was not time limited).
 * 5. Unlike the basic "jdemo.java", this demo processes only the first key
 *    in its input file.  Therefore, provide a file containing only one key.
 *    The key need not be quota limited.  If it isn't, the key cookie will
 *    still manage a use count, which can be useful in an actual application.
 */


import java.io.*;
import java.text.*;
import java.util.*;
import com.vs.ezlicrun.*;

public class jdemoextapi {

	static final int KEY_MAXLENGTH = 4096;

	static String product = null;
	static String appPwd = null;
	static long quotaIncr = 0;
	static String command = null;
	static String userhost = null;
	static String auxkeyfilename = null;
    static String keyfilename = null;
	static String hfileDir = null;

	public jdemoextapi() {}

	static void reportParams() {
		System.err.println("Input Parameters:");
		System.err.println("  Command: " + command);
		System.err.println("  Product name: " + product);
		System.err.println("  Application password public key: " + (appPwd == null? "<<not specified>>": appPwd));
		System.err.println("  Logical user / host name in license key: " + (userhost == null ? "<<not specified>>": userhost));
		System.err.println("  License key / cookie input file: " + keyfilename);
		System.err.println("  Auxiliary Agilis license key input file: " + auxkeyfilename);
		System.err.println("  Metered-license quota consumption increment: " + (quotaIncr <= 0? "<<not specified>>": String.valueOf(quotaIncr)));
		System.err.println("  Hidden file directory: " + (hfileDir == null? "<<not specified>>": hfileDir));
	}

	static void printUsage(String msg) {
		if (msg != null)
			System.err.println("*** " + msg + " ***");
		System.err.println("Usage: jdemoextapi -cCommand -pProduct [-wPassword] [-uUserhost] [-fLicenseKeyOrCookieFile] [-aAuxLicensefile] [-qQuotaIncrement] [-rHiddenFileDir]");
		System.err.println("  -cCommand: one of 'activate', 'check' or 'deactivate'.  Required.");
		System.err.println("  -pProduct: name of product for which license key / key cookie is defined.  Required.");
		System.err.println("  -wPassword: application password public key.  Optional, defaults to null.");
		System.err.println("  -uUserhost: logical user/host name matching license key.  Optional, defaults to pre-node-locked.");
		System.err.println("  -fLicenseKeyOrCookieFile: name of file containing license key or cookie.  Optional, defaults to stdin.");
		System.err.println("  -aAuxLicenseFile: name of file containing auxiliary Agilis license key for the product.  Optional.");
		System.err.println("  -qQuotaIncrement: amount by which to increment quota consumption.  Optional, defaults to 1.");
		System.err.println("  -rHiddenFileDir: hidden file directory.  Optional, defaults to current working directory.");
		System.exit(2);
	}

	static void setDefaults() {
		if (quotaIncr <= 0)
			quotaIncr = 1;
		if (hfileDir == null)
			hfileDir = ".";
	}

	static void getArgs(String[] args) {
		if (args.length <= 2)
			printUsage("Insufficient arguments");

		String arg = null;
		for (int i = 0; i < args.length; i++) {
			arg = args[i];
			if (arg.startsWith("-p")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -p ***");
					System.exit(1);
				}
				product = arg.substring(2);
			} else if (arg.startsWith("-c")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -c ***");
					System.exit(1);
				}
				command = arg.substring(2);
				if (!(command.equalsIgnoreCase("activate") ||
					  command.equalsIgnoreCase("check") ||
					  command.equalsIgnoreCase("deactivate")))
					printUsage("Invalid command code");
			} else if (arg.startsWith("-w")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -w ***");
					System.exit(1);
				}
				appPwd = arg.substring(2);
			} else if (arg.startsWith("-u")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -u ***");
					System.exit(1);
				}
				userhost = arg.substring(2);
			} else if (arg.startsWith("-f")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -a ***");
					System.exit(1);
				}
				keyfilename = arg.substring(2);
			} else if (arg.startsWith("-a")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -a ***");
					System.exit(1);
				}
				auxkeyfilename = arg.substring(2);
			} else if (arg.startsWith("-q")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -q ***");
					System.exit(1);
				}
				try {
					quotaIncr = Long.parseLong(arg.substring(2));
				} catch (Throwable t) {
					quotaIncr = -1;
				}
				if (quotaIncr <= 0) {
					System.err.println("*** Quota increment must be a positive integer ***");
					System.exit(1);
				}
			} else if (arg.startsWith("-r")) {
				if (arg.length() == 2) {
					System.err.println("*** No argument specified for -h ***");
					System.exit(1);
				}
				hfileDir = arg.substring(2);
				File f = new File(hfileDir);
				if (!f.exists())
					printUsage("Hidden file directory '" + hfileDir + "' does not exist");
			} else {
				String msg = null;
				if (!arg.startsWith("-h"))
					msg = "Invalid argument: '" + arg + "'";
				printUsage(msg);
			}
		}
		if (product == null || command == null)
			printUsage("Missing mandatory parameters");
	}

	static String readKey(String filename) {
		String keytext = "";
		BufferedReader in = null;
		try {
			in = new BufferedReader(
					new InputStreamReader(
						filename == null ? System.in : new FileInputStream(filename)));
			String keyline = null;
			while ((keyline = in.readLine()) != null)
				keytext += keyline;
		} catch (Throwable t) {
			System.err.println("*** IO Error reading license key file '" + (filename == null ? "stdin":filename) + " ***");
			System.err.println("Exception: " + t.getClass().getName() + ", Message: " + t.getMessage());
			t.printStackTrace();
			System.exit(2);
		} finally {
			try {
				in.close();
			} catch (Throwable tt) {
			}
		}
		keytext = strip(keytext);
		return keytext;
	}

	static String strip(String instr) {
		String retstr = "";
		try {
			StringTokenizer stk = new StringTokenizer(instr," \n\t");
			while (stk.hasMoreTokens()) {
				String tokstr = stk.nextToken();
				retstr += tokstr;
			}
		} catch (Exception ex) {
			System.out.println("*** Got exception: " + ex.getMessage());
		}
		return retstr;
	}
	
	/**
	 * main entry point:
	 * Accepts mandatory command line arguments:
	 * 1. Case-sensitive product name that was specified at the time of key creation.
	 *    This is required even if the product does not have an associated 
	 *    application password.
	 * 2. Command: activate / check / deactivate
	 * and the optional arguments
	 * 3. Name of a text file containing a license key to be activated
	 *    if invoked in activation mode, or containing a key cookie in
	 *    .key notation if invoked in license check or deactivation mode.
	 *    If not specified, the license or key cookie is read from
	 *    the standard input stream, in which case it is expected to
	 *    be just the key with no additional formatting information.
 	 * 4. Quota increment: amount by which to increment quota consumption.
 	 *    To reset the saved state, specify a value of 0 to simulate
 	 *    product install time.
	 * 5. Expected user / host name that should match what's in the key.
     *    If the key is known to have been generated with the "enforce OS
     *    user / host" flag, you don't need to specify this parameter.
     *    Also, if invoked in deferred node locking mode, this name does 
     *    not have to be the OS user/host name.
	 * 6. Product's associated password's case-sensitive public key, available with
	 *    the product definition and also obtainable by invoking PkCrypt.makePublicKey(password)
	 *    and with the "makepubkey" Java utility located at ezlic2.0/util.
	 * 7. Auxiliary Agilis key, required if deferred time-limited licensing and deferred
	 *    node locking functionality are to be used.  This is a special key provided to the vendor
	 *    by Agilis Software.  It cannot be generated by the vendor.
	 * When invoked in activation or license check modes, the newly-generated key cookie is
	 * outputted to the standard-out stream, which may be redirected to a disk file in the
	 * command line invocation.  All messages produced by the demo are outputted to standard
	 * error.
	 */
	public static void main(String[] args) {
		boolean ret = false;
		getArgs(args);
		System.err.println("EasyLicenser Extended API Demo.");
		System.err.println("Running on " + (new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).format(new Date()));
		reportParams();
		setDefaults();
		String keytxt = readKey(keyfilename);
		String auxkeytxt = auxkeyfilename == null ? null :
						readKey(auxkeyfilename);
		if (command.equalsIgnoreCase("activate"))
			doActivate(keytxt, auxkeytxt);
		else if (command.equalsIgnoreCase("check"))
			doCheck(keytxt);
		else if (command.equalsIgnoreCase("deactivate"))
			doDeactivate(keytxt);
		else {
			System.err.println("*** Unrecognized command: '" + command + "' ***");
			System.exit(2);
		}
		System.err.println("End demo.");
		System.exit(0);
	}

	static void doActivate(String keytxt, String auxkeytxt) {
		EzLicenseExtendedAPI licInfo = null;
		System.err.println("Activating license...");
		try {
			// EasyLicenser Extended API call: activate license, possibly performing
			// a deferred node lock and initializing deferred time limited licensing
			// if an appropriate auxiliary Agilis license key is provided.
			licInfo = 
				EzLicenseExtendedAPI.activateLicense(
					keytxt, 		// actual license key text
					auxkeytxt,		// auxiliary Agilis license key text
					null,			// no custom key handler in this example
					null,			// no custom key context in this example
					userhost,		// logical user/host name (does not have to be
									// OS-level user/host if using deferred node
									// locking.
					product,		// case-sensitive product name
					appPwd,			// product application password public key
					"userappstateexample", // user-specified application state
					hfileDir,	// hidden file directory
					false,			// check system clock only if time limited license
					auxkeytxt != null // deferred node locking only if 
									// auxiliary key specified
					);
		} catch (EzLicExceptionBase ex) {
			System.err.println("*** Error activating license.  Exception: " + ex.getClass().getName() +
							   ", message: " + ex.getMessage());
			System.exit(2);
		}
		System.err.println("Activated license.");
		checkWarnings(licInfo.getStatusCode());
		displayLicenseParams(licInfo);
		displayExtendedLicenseState(licInfo);
		System.out.println(licInfo.getKeyCookie());
	}

	static void doCheck(String keytxt) {
		EzLicenseExtendedAPI licInfo = null;
		System.err.println("Check license...");
		try {
			// EasyLicenser Extended API call: check license, possibly performing
			// checks for deferred node locking and deferred time limited licensing
			// if an appropriate auxiliary Agilis license key was provided at
			// the time of activation.
			licInfo = 
				EzLicenseExtendedAPI.checkLicense(
					keytxt, 		// actual license key cookie text
					null,			// no custom key handler in this example
					null,			// no custom key context in this example
					5,				// days-left warning threshold
					-quotaIncr,		// quota usage to date: -ve value implies
									// quota to deduct as for secure API calls
					10,				// quota warning threshold
					0,				// past-expiration days grace
					0,				// over-quota grace
					userhost,		// logical user/host name (does not have to be
									// OS-level user/host if using deferred node
									// locking.
					0,				// current usage level - not interesting for
									// this demo
					product,		// case-sensitive product name
					appPwd,			// product application password public key
					"userappstateexample2", // user-specified application state
					hfileDir,	// hidden file directory
					false			// check system clock only if time limited license
					);
		} catch (EzLicExceptionBase ex) {
			System.err.println("*** Error checking license.  Exception: " + ex.getClass().getName() +
							   ", message: " + ex.getMessage());
			System.exit(2);
		}
		System.err.println("Checked license.");
		checkWarnings(licInfo.getStatusCode());
		displayLicenseParams(licInfo);
		displayExtendedLicenseState(licInfo);
		System.out.println(licInfo.getKeyCookie());
	}

	static void doDeactivate(String keytxt) {
		EzLicenseExtendedAPI licInfo = null;
		System.err.println("Deactivating license...");
		try {
			// EasyLicenser Extended API call: deactivate license, after
			// first performing a lenient check on it to ensure it is
			// a valid extended-API key cookie for the product.
			EzLicenseExtendedAPI.deactivateLicense(keytxt, product, appPwd, hfileDir, true);
		} catch (EzLicExceptionBase ex) {
			System.err.println("*** Error checking license.  Exception: " + ex.getClass().getName() +
							   ", message: " + ex.getMessage());
			System.exit(2);
		}
		System.err.println("Deactivated license.");
	}

	/*
	 * Display state contents of extended-license contained in key cookie.
	 * Note that you can introspect a key cookie but you can't create /
	 * update it outside the EasyLicenser extended API.  In an actual
	 * application, the only item that may be interesting to the
	 * application would be the application-specified state information
	 * provided at the time of an activation or license check.
	 */
	static void displayExtendedLicenseState(EzLicenseExtendedAPI info) {
		System.err.println("Extended License State:");
		System.err.println("(a) Transient state from last operation:");
		System.err.println("    Key cookie: " + info.getKeyCookie());
		System.err.println("    Operation status code: " + info.getStatusCode());
		System.err.println("(b) Persistent state:");
		System.err.println("    Original license key (first few characters):\n\t" + info.getLicKey().substring(0, 20) + "...");
		System.err.println("    User application state: " + info.getUserAppstate());
		System.err.println("    OS User/host name: " + info.getUserhost());
		System.err.println("    Activation date/time: " +
							(new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).format(new Date(info.getActTs())));
		if (info.getExpTs() > 0)
			System.err.println("    Expiration date/time: " +
							(new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).format(new Date(info.getExpTs())));
		else
			System.err.println("    Not a deferred-time-limited license.");
		System.err.println("(c) Key cookie internal state:");
		String keyCookie = info.getKeyCookie();
		try {
			/* EasyLicenser API call: get application state from key cookie */
//			String appState = 
//				EzLicenseInfo.
//					getKeyCookieAppState(keyCookie, product, appPwd);
//			System.err.println("    Application state: [" + appState + "]");

			/* EasyLicenser API call: get quota-consumed-to-date state from key cookie */
			long quotaToDate = 
				EzLicenseInfo.
					getKeyCookieQuotaToDate(keyCookie, product, appPwd);
			System.err.println("    Quota-to-date: " + quotaToDate);

			/* EasyLicenser API call: get cookie update timestamp from key cookie */
			long updateTs = 
				EzLicenseInfo.
					getKeyCookieTs(keyCookie, product, appPwd);
			System.err.println("    Update date/time: " +
						(new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).
						format(new Date(updateTs)) + " (timestamp: " +
						updateTs + ")");
		} catch (EzLicExceptionBase ex) {
			System.out.println("*** Error introspecting key cookie.  Exception: " +
				ex.getClass().getName() + ", Message: " +
				ex.getMessage());
			System.exit(2);
		}
	}

	static void checkWarnings(int warnings)
	{
		if (warnings != 0) {
			System.err.println("    Warnings:");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRING) != 0)
				System.err.println("        Time Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA) != 0)
				System.err.println("        Quota Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRED) != 0)
				System.err.println("        Time Limited Key Has Expired");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA_EXCEEDED) != 0)
				System.err.println("        Quota Limited Key Has Expired");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXP_GRACE_TRUNC) != 0)
				System.err.println("        Expiration Days Grace In API Call Too Large - Truncated To 5 / 25% Of Lifespan");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA_GRACE_TRUNC) != 0)
				System.err.println("        Quota Grace In API Call Too Large - Truncated To 5 / 25% Of Original Quota");
			System.err.println("");
		}
	}

	static void displayLicenseParams(EzLicenseInfo licenseInfo)
	{
		System.err.println("    License Key Information:");
		System.err.println("        Key Version: " + 
							(licenseInfo.getKeyVersion() <= 0 ? "Legacy (1.1 or earlier)" :
							 String.valueOf(licenseInfo.getKeyVersion()/10)+"." + 
							 String.valueOf(licenseInfo.getKeyVersion()%10)
							));
		if (licenseInfo.getChainedKeyId() != null)
			System.err.println("Replacement For Key With Id: " + licenseInfo.getChainedKeyId());
		showLicenseModel(licenseInfo);
		int bitmap = licenseInfo.getLicenseModeBitmap();
		System.err.println("        License Mode Bitmap: " + bitmap);
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_TIME) != 0)
			System.err.println("            TIME LIMITED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_OPTIONS) != 0)
			System.err.println("            OPTIONS");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_METERED) != 0)
			System.err.println("            METERED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_KEY) != 0)
			System.err.println("            CUSTOM KEY");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_COOKIE) != 0)
			System.err.println("            CUSTOM COOKIE");
		long lvalue;
		lvalue = licenseInfo.getUsageValue();
		if (lvalue > 0)
			System.err.println("        Usage Limit Value: " + lvalue);
		lvalue = licenseInfo.getQuotaValue();
		if (lvalue > 0)
			System.err.println("        Quota Value: " + lvalue);
		String options = licenseInfo.getOptions();
		if (options != null && !options.equals(""))
			System.err.println("        Options: [" + options + "]");
		Date expDate = licenseInfo.getExpirationDate();
		if (expDate != null)
			System.err.println("        Expiration Date: " +
							(new SimpleDateFormat("yyyy-MM-dd hh:mm:ss z")).
							format(expDate));
		System.err.println("        User / host name: " +
						licenseInfo.getUserHostNetName());
		if (licenseInfo.getEnforce())
			System.err.println("        Operating system check enforced.");
		String customCookie = licenseInfo.getCustomCookie();
		if (customCookie != null && !customCookie.equals(""))
			System.err.println("        Custom Cookie: [" + customCookie + "]");
		String customKey = licenseInfo.getCustomKey();
		if (customKey != null && !customKey.equals(""))
			System.err.println("        Custom Key: [" + customKey + "]");
		if (licenseInfo.getAppPwd() != null && !licenseInfo.getAppPwd().equals(""))
			System.err.println("        Application Password: [" + licenseInfo.getAppPwd() + "]");
	}

	static void showLicenseModel(EzLicenseInfo licenseInfo)
	{
		int licenseModel = licenseInfo.getLicenseModelCode();
		int licenseType = licenseInfo.getLicenseTypeCode();
		System.err.println("        License Model: " +
						   (licenseModel == EzLicenseInfo.EZLIC_MODEL_USER?"User":"Server"));
		System.err.print("        License Type: ");
		switch(licenseModel) {
		case EzLicenseInfo.EZLIC_MODEL_USER:
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_U_USER:
				System.err.println("Named User");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_NODE:
				System.err.println("Node Locked");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_FLOAT:
				System.err.println("Floating User");
				break;
			default:
				System.err.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		case EzLicenseInfo.EZLIC_MODEL_SERVER:
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_SVR_CONC:
				System.err.println("Concurrent User");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_CPU:
				System.err.println("CPU Count");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_MHZ:
				System.err.println("CPU MHZ");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_NMU:
				System.err.println("Named User");
				break;
			default:
				System.err.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		default:
			System.err.println("*** Unknown license model " + licenseModel + " ***");
		}
	}

}
